@@ -28,6 +28,7 @@ from abc import ABC, abstractmethod
 from collections import deque 
 from itertools import chain
 import random
+import numpy as np 
 import jax 
 from jax import numpy as jnp
 
@@ -36,35 +37,74 @@ from .episode_tracer import Transition
 
    
 class Trajectory:
+    r"""
+    A simple trajectory to hold episodical transitions.
+    """
     def __init__(self):
       self.trajectory = sliceable_deque([])
       self._transition_weight = sliceable_deque([])
+      self._batched_transitions = None
     
     def add(self, transition):
+      """Adds a transition to the trajectory
+      
+      Parameters
+      ----------
+      transition: An instance of muax.episode_tracer.Transition. 
+      
+      """
       self.trajectory.append(transition)
       self._transition_weight.append(transition.w)
 
     
+    def finalize(self):
+      """"""
+      batched_transitions = jax.tree_util.tree_transpose(
+          outer_treedef=jax.tree_util.tree_structure([0 for i in self.trajectory]),
+          inner_treedef=jax.tree_util.tree_structure(Transition()),
+          pytree_to_transpose=self.trajectory
+          )
+      batched_transitions = Transition(*(np.expand_dims(np.vstack(_attr), axis=0)
+          for _attr in batched_transitions))
+      self._batched_transitions = batched_transitions
+
+    
     def sample(self, num_samples: int = 1, k_steps: int = 5):
+      """Samples consecutive transitions of k steps from the trajectory.
+      
+      Parameters
+      ----------
+      num_samples: int, positive int. Number of samples to draw from the trajectory. 
+          Each sample is chosen given the weight(transition.w) as probability.
+      k_steps: int, positive int. Determines the length of consecutive steps in each sample.
+          If k_steps >= len(trajectory), no samples would be collected.
+      
+      Returns
+      ----------
+      samples: List[Transition]. For each transition in the list, 
+          the attributes of which have the dimension `[B, L, ...]`, where B is the batch size(1) and 
+          L is the length(k) of the consecutive transitions.
+      """
       if len(self) <= k_steps: return []
       max_idx = len(self) - k_steps
       idxes = random.choices(range(max_idx), 
-                             weights=self._transition_weight[:max_idx], 
-                             k=num_samples)
-      
+                            weights=self._transition_weight[:max_idx], 
+                            k=num_samples)
+                            
+      if self.batched_transitions is None:
+        self.finalize()
+
       samples = [self._get_sample(idx, k_steps) for idx in idxes]
       
       return samples
 
+    @property
+    def batched_transitions(self):
+      return self._batched_transitions
+
     def _get_sample(self, idx, k_steps):
       end_idx = idx + k_steps 
-      k_steps_sample = self[idx: end_idx]
-      sample = jax.tree_util.tree_transpose(
-          outer_treedef=jax.tree_util.tree_structure([0 for i in k_steps_sample]),
-          inner_treedef=jax.tree_util.tree_structure(Transition()),
-          pytree_to_transpose=k_steps_sample
-          )
-      sample = Transition(*(jnp.expand_dims(jnp.vstack(_attr), axis=0) for _attr in sample))
+      sample = self.batched_transitions[:, idx: end_idx, :]
       return sample
 
     def __getitem__(self, index):
@@ -74,7 +114,7 @@ class Trajectory:
       return len(self.trajectory)
 
     def __repr__(self):
-      return f'{type(self)(len={len(self)})}'
+      return f'{type(self)}(len={len(self)})'
       
 
 class BaseReplayBuffer(ABC):
@@ -167,7 +207,7 @@ class TrajectoryReplayBuffer(BaseReplayBuffer):
         """
         if batch_size is None and num_trajectory is None: 
           raise ValueError('Either num_trajectory or batch_size need to be given.')
-        if batch_size is not None:
+        elif batch_size is not None and num_trajectory is None:
           num_trajectory = batch_size 
           sample_per_trajectory = 1
         # sandwich sample in between setstate/getstate in case global random state was tampered with
@@ -184,7 +224,7 @@ class TrajectoryReplayBuffer(BaseReplayBuffer):
           outer_treedef=jax.tree_util.tree_structure([0 for i in batch]),
           inner_treedef=jax.tree_util.tree_structure(Transition()),
           pytree_to_transpose=batch)
-        batch = Transition(*(jnp.vstack(v) for v in batch))
+        batch = Transition(*(np.vstack(v) for v in batch))
         return batch
 
     def clear(self):
